#ifndef CONFIG
#define CONFIG

#define VERSION             220

//try to keep this in sync.
//SavvyCAN will complain if you connect a GVRET board with a revision
//less than this number.
#define CURRENT_GVRET_VER   343

#endif // CONFIG

